"""Clinical file format validation and processing"""
import datetime
from io import StringIO
import logging
import os

import pandas as pd
import synapseclient

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions
from genie.database_to_staging import redact_phi

logger = logging.getLogger(__name__)


# def remove_greaterthan_lessthan_str(col):
#     '''
#     In clinical file, there are redacted value such as >89 and <17.
#     These < and > signs must be removed
#     '''
#     try:
#         col = [
#             text.replace(">", "")
#             if isinstance(text, str) else text for text in col]
#         col = [
#             int(text.replace("<", ""))
#             if isinstance(text, str) and text != "" else text
#             for text in col]
#     except ValueError:
#         pass
#     return(col)
def _check_year(clinicaldf: pd.DataFrame, year_col: int, filename: str,
                allowed_string_values: list = []) -> str:
    """Check year columns

    Args:
        clinicaldf: Clinical dataframe
        year_col: YEAR column
        filename: Name of file
        allowed_string_values: list of other allowed string values

    Returns:
        Error message
    """
    error = ''
    if process_functions.checkColExist(clinicaldf, year_col):
        # Deal with pre-redacted values and other allowed strings
        # first because can't int(text) because there are
        # instances that have <YYYY
        year_series = clinicaldf[year_col][
            ~clinicaldf[year_col].isin(allowed_string_values)
        ]
        year_now = datetime.datetime.utcnow().year
        try:
            years = year_series.apply(
                lambda x: datetime.datetime.strptime(
                    str(int(x)), '%Y').year > year_now
            )
            # Make sure that none of the years are greater than the current
            # year.  It can be the same, but can't future years.
            assert not years.any()
        except Exception:
            error = (f"{filename}: Please double check your {year_col} "
                     "column, it must be an integer in YYYY format "
                     f"<= {year_now}")
            # Tack on allowed string values
            if allowed_string_values:
                error += " or '{}'.\n".format(
                    "', '".join(allowed_string_values)
                )
            else:
                error += ".\n"
    else:
        error = f"{filename}: Must have {year_col} column.\n"

    return error


# PROCESSING
def remap_clinical_values(clinicaldf: pd.DataFrame, sex_mapping: pd.DataFrame,
                          race_mapping: pd.DataFrame,
                          ethnicity_mapping: pd.DataFrame,
                          sampletype_mapping: pd.DataFrame) -> pd.DataFrame:
    """Remap clinical attributes from integer to string values

    Args:
        clinicaldf: Clinical data
        sex_mapping: Sex mapping data
        race_mapping: Race mapping data
        ethnicity_mapping: Ethnicity mapping data
        sample_type: Sample type mapping data

    Returns:
        Mapped clinical dataframe
    """

    race_mapping.index = race_mapping['CODE']
    race_dict = race_mapping.to_dict()

    ethnicity_mapping.index = ethnicity_mapping['CODE']
    ethnicity_dict = ethnicity_mapping.to_dict()

    sex_mapping.index = sex_mapping['CODE']
    sex_dict = sex_mapping.to_dict()

    sampletype_mapping.index = sampletype_mapping['CODE']
    sampletype_dict = sampletype_mapping.to_dict()

    if clinicaldf.get("SAMPLE_TYPE") is not None:
        clinicaldf['SAMPLE_TYPE_DETAILED'] = clinicaldf['SAMPLE_TYPE']

    # Use pandas mapping feature
    clinicaldf = clinicaldf.replace({
        "PRIMARY_RACE": race_dict['CBIO_LABEL'],
        "SECONDARY_RACE": race_dict['CBIO_LABEL'],
        "TERTIARY_RACE": race_dict['CBIO_LABEL'],
        "SAMPLE_TYPE": sampletype_dict['CBIO_LABEL'],
        "SAMPLE_TYPE_DETAILED": sampletype_dict['DESCRIPTION'],
        "SEX": sex_dict['CBIO_LABEL'],
        'ETHNICITY': ethnicity_dict['CBIO_LABEL']
    })

    return clinicaldf


class clinical(FileTypeFormat):

    _fileType = "clinical"

    # _process_kwargs = [
    #     "newPath", "patientSynId", "sampleSynId",
    #     "parentId", "retractedSampleSynId", "retractedPatientSynId"]
    _process_kwargs = [
        "newPath", "parentId", "databaseToSynIdMappingDf", "oncotree_link",
        'clinicalTemplate', 'sample', 'patient', 'patientCols', 'sampleCols']

    _validation_kwargs = ["oncotree_link"]

    # VALIDATE FILE NAME
    def _validateFilename(self, filePath):
        if len(filePath) == 1:
            assert os.path.basename(filePath[0]) == \
                "data_clinical_supp_{}.txt".format(self.center)
        else:
            required = pd.Series([
                "data_clinical_supp_sample_{}.txt".format(self.center),
                "data_clinical_supp_patient_{}.txt".format(self.center)])
            assert all(required.isin([os.path.basename(i) for i in filePath]))

    # Update clinical file with the correct mappings
    def update_clinical(self, row):
        """Transform the values of each row of the clinical file"""
        # Must create copy or else it will overwrite the original row
        x = row.copy()
        # # PATIENT ID
        if x.get("PATIENT_ID") is not None:
            x['PATIENT_ID'] = process_functions.checkGenieId(
                x['PATIENT_ID'], self.center)
        # # RACE
        if x.get('PRIMARY_RACE') is None:
            x['PRIMARY_RACE'] = "Not Collected"

        if x.get('SECONDARY_RACE') is None:
            x['SECONDARY_RACE'] = "Not Collected"

        if x.get('TERTIARY_RACE') is None:
            x['TERTIARY_RACE'] = "Not Collected"
        # ETHNICITY
        if x.get('ETHNICITY') is None:
            x['ETHNICITY'] = "Not Collected"
        # BIRTH YEAR
        if x.get("BIRTH_YEAR") is not None:
            # BIRTH YEAR (Check if integer)
            if process_functions.checkInt(x['BIRTH_YEAR']):
                x['BIRTH_YEAR'] = int(x['BIRTH_YEAR'])

        # SAMPLE ID
        if x.get('SAMPLE_ID') is not None:
            x['SAMPLE_ID'] = process_functions.checkGenieId(
                x['SAMPLE_ID'], self.center)

        # AGE AT SEQ REPORT
        if x.get('AGE_AT_SEQ_REPORT') is not None:
            if process_functions.checkInt(x['AGE_AT_SEQ_REPORT']):
                x['AGE_AT_SEQ_REPORT'] = int(x['AGE_AT_SEQ_REPORT'])

        # SEQ ASSAY ID
        if x.get('SEQ_ASSAY_ID') is not None:
            x['SEQ_ASSAY_ID'] = x['SEQ_ASSAY_ID'].replace('_', '-')
            # standardize all SEQ_ASSAY_ID with uppercase
            x['SEQ_ASSAY_ID'] = x['SEQ_ASSAY_ID'].upper()

        if x.get('SEQ_DATE') is not None:
            x['SEQ_DATE'] = x['SEQ_DATE'].title()
            x['SEQ_YEAR'] = \
                int(str(x['SEQ_DATE']).split("-")[1]) \
                if str(x['SEQ_DATE']) != "Release" else float('nan')

        if x.get('YEAR_CONTACT') is None:
            x['YEAR_CONTACT'] = 'Not Collected'
        else:
            if process_functions.checkInt(x['YEAR_CONTACT']):
                x['YEAR_CONTACT'] = int(x['YEAR_CONTACT'])

        if x.get('YEAR_DEATH') is None:
            x['YEAR_DEATH'] = 'Not Collected'
        else:
            if process_functions.checkInt(x['YEAR_DEATH']):
                x['YEAR_DEATH'] = int(x['YEAR_DEATH'])

        if x.get('INT_CONTACT') is None:
            x['INT_CONTACT'] = 'Not Collected'

        if x.get('INT_DOD') is None:
            x['INT_DOD'] = 'Not Collected'

        if x.get('DEAD') is None:
            x['DEAD'] = 'Not Collected'

        # TRIM EVERY COLUMN MAKE ALL DASHES
        for i in x.keys():
            if isinstance(x[i], str):
                x[i] = x[i].strip(" ")
        return x

    def uploadMissingData(self, df, col, dbSynId, stagingSynId,
                          retractionSynId=None):
        """Uploads missing clinical samples / patients"""
        samples = "','".join(df[col])
        path = os.path.join(
            process_functions.SCRIPT_DIR,
            f"{self._fileType}_missing_{col}.csv"
        )
        missing = self.syn.tableQuery(
            f"select {col} from {dbSynId} where "
            f"CENTER='{self.center}' and {col} not in ('{samples}')"
        )
        missing.asDataFrame().to_csv(path, index=False)
        self.syn.store(synapseclient.File(path, parent=stagingSynId))
        os.remove(path)

    def _process(self, clinical, clinicalTemplate):
        # Capitalize all clinical dataframe columns
        clinical.columns = [col.upper() for col in clinical.columns]
        clinical = clinical.fillna("")
        # clinicalMerged = clinical.merge(clinicalTemplate,how='outer')
        # Remove unwanted clinical columns prior to update
        # clinicalMerged = clinicalMerged.drop(clinicalMerged.columns[
        #    ~clinicalMerged.columns.isin(clinicalTemplate.columns)],1)
        ethnicity_mapping = process_functions.getGenieMapping(
            self.syn, "syn7434242")
        race_mapping = process_functions.getGenieMapping(
            self.syn, "syn7434236")
        sex_mapping = process_functions.getGenieMapping(self.syn, "syn7434222")
        sampletype_mapping = process_functions.getGenieMapping(
            self.syn, "syn7434273")
        # Attach MSK to centers
        # clinicalMerged = clinicalMerged.fillna("")
        clinical = remap_clinical_values(
            clinical, sex_mapping, race_mapping, ethnicity_mapping,
            sampletype_mapping
        )
        remapped_clindf = clinical.apply(self.update_clinical, axis=1)
        # Some columns may have been added during update,
        # remove unwanted columns again
        keep_cols_idx = remapped_clindf.columns.isin(clinicalTemplate.columns)
        remapped_clindf = remapped_clindf.drop(
            remapped_clindf.columns[~keep_cols_idx], 1
        )
        remapped_clindf['CENTER'] = self.center
        return remapped_clindf

    def preprocess(self, newpath):
        """
        Gather preprocess parameters

        Args:
            filePath: Path to file

        Returns:
            dict with keys - 'clinicalTemplate', 'sample', 'patient',
                             'patientCols', 'sampleCols'
        """
        # These synapse ids for the clinical tier release scope is
        # hardcoded because it never changes
        patient_cols_table = self.syn.tableQuery(
            'select fieldName from syn8545211 where '
            'patient is True and inClinicalDb is True'
        )
        patient_cols = patient_cols_table.asDataFrame()['fieldName'].tolist()
        sample_cols_table = self.syn.tableQuery(
            'select fieldName from syn8545211 where '
            'sample is True and inClinicalDb is True'
        )
        sample_cols = sample_cols_table.asDataFrame()['fieldName'].tolist()
        clinicalTemplate = pd.DataFrame(
            columns=set(patient_cols + sample_cols)
        )
        sample = True
        patient = True

        return({'clinicalTemplate': clinicalTemplate,
                'sample': sample,
                'patient': patient,
                'patientCols': patient_cols,
                'sampleCols': sample_cols})

    def process_steps(self, clinicalDf,
                      databaseToSynIdMappingDf, newPath,
                      parentId, oncotree_link, clinicalTemplate,
                      sample, patient, patientCols, sampleCols):
        """Process clincial file, redact PHI values, upload to clinical
        database
        """
        patientdb_idx = databaseToSynIdMappingDf['Database'] == "patient"
        patient_synid = databaseToSynIdMappingDf.Id[patientdb_idx][0]
        sampledb_idx = databaseToSynIdMappingDf['Database'] == "sample"
        sample_synid = databaseToSynIdMappingDf.Id[sampledb_idx][0]

        newClinicalDf = self._process(clinicalDf, clinicalTemplate)
        newClinicalDf = redact_phi(newClinicalDf)

        if patient:
            patientClinical = newClinicalDf[
                patientCols].drop_duplicates("PATIENT_ID")
            self.uploadMissingData(patientClinical, "PATIENT_ID",
                                   patient_synid, parentId)

            process_functions.updateData(self.syn, patient_synid,
                                         patientClinical, self.center,
                                         col=patientCols, toDelete=True)
        if sample:
            if sum(newClinicalDf["SAMPLE_ID"].duplicated()) > 0:
                logger.error("There are duplicated samples, "
                             "and the duplicates are removed")
            sampleClinical = newClinicalDf[sampleCols].drop_duplicates(
                "SAMPLE_ID")
            # Exclude all clinical samples with wrong oncotree codes
            oncotree_mapping = pd.DataFrame()
            oncotree_mapping_dict = \
                process_functions.get_oncotree_code_mappings(oncotree_link)
            # Add in unknown key for oncotree code
            oncotree_mapping_dict['UNKNOWN'] = {}
            oncotree_mapping['ONCOTREE_CODE'] = list(oncotree_mapping_dict.keys())
            # Make oncotree codes uppercase (SpCC/SPCC)
            sampleClinical['ONCOTREE_CODE'] = sampleClinical[
                'ONCOTREE_CODE'].astype(str).str.upper()
            sampleClinical = sampleClinical[sampleClinical[
                'ONCOTREE_CODE'].isin(oncotree_mapping['ONCOTREE_CODE'])]
            self.uploadMissingData(
                sampleClinical, "SAMPLE_ID", sample_synid, parentId)
            # ,retractedSampleSynId)
            process_functions.updateData(
                self.syn, sample_synid, sampleClinical,
                self.center, col=sampleCols, toDelete=True)

        newClinicalDf.to_csv(newPath, sep="\t", index=False)
        return(newPath)

    # VALIDATION
    def _validate(self, clinicaldf, oncotree_link):
        """
        This function validates the clinical file to make sure it adhere
        to the clinical SOP.

        Args:
            clinicalDF: Merged clinical file with patient and sample
                        information
            oncotree_link: Link to oncotree

        Returns:
            Error message
        """
        total_error = StringIO()
        warning = StringIO()

        clinicaldf.columns = [col.upper() for col in clinicaldf.columns]
        clinicaldf = clinicaldf.fillna("")

        # oncotree_mapping = process_functions.get_oncotree_codes(oncotree_link)
        # if oncotree_mapping.empty:
        oncotree_mapping_dict = \
            process_functions.get_oncotree_code_mappings(oncotree_link)
        oncotree_mapping = pd.DataFrame(
            {"ONCOTREE_CODE": list(oncotree_mapping_dict.keys())}
        )

        sampletype_mapping = \
            process_functions.getGenieMapping(self.syn, "syn7434273")

        ethnicity_mapping = \
            process_functions.getGenieMapping(self.syn, "syn7434242")

        race_mapping = \
            process_functions.getGenieMapping(self.syn, "syn7434236")

        sex_mapping = \
            process_functions.getGenieMapping(self.syn, "syn7434222")

        # CHECK: SAMPLE_ID
        sampleId = 'SAMPLE_ID'
        haveSampleColumn = \
            process_functions.checkColExist(clinicaldf, sampleId)

        if not haveSampleColumn:
            total_error.write(
                "Sample Clinical File: Must have SAMPLE_ID column.\n"
            )
        else:
            if sum(clinicaldf[sampleId].duplicated()) > 0:
                total_error.write(
                    "Sample Clinical File: No duplicated SAMPLE_ID "
                    "allowed.\nIf there are no duplicated "
                    "SAMPLE_IDs, and both sample and patient files are "
                    "uploaded, then please check to make sure no duplicated "
                    "PATIENT_IDs exist in the patient clinical file.\n"
                )
        # CHECK: PATIENT_ID
        patientId = "PATIENT_ID"
        # #CHECK: PATIENT_ID IN SAMPLE FILE
        havePatientColumn = \
            process_functions.checkColExist(clinicaldf, patientId)

        if not havePatientColumn:
            total_error.write(
                "Patient Clinical File: Must have PATIENT_ID column.\n"
            )

        # CHECK: within the sample file that the sample ids match
        # the patient ids
        if haveSampleColumn and havePatientColumn:
            # Make sure sample and patient ids are string cols
            clinicaldf[sampleId] = clinicaldf[sampleId].astype(str)
            clinicaldf[patientId] = clinicaldf[patientId].astype(str)
            if not all([patient in sample
                        for sample, patient in
                        zip(clinicaldf[sampleId], clinicaldf[patientId])]):

                total_error.write(
                    "Sample Clinical File: PATIENT_ID's much be contained in "
                    "the SAMPLE_ID's (ex. SAGE-1 <-> SAGE-1-2)\n"
                )
            # #CHECK: All samples must have associated patient data
            # (GENIE requires patient data)
            if not all(clinicaldf[patientId] != ""):
                total_error.write(
                    "Patient Clinical File: All samples must have associated "
                    "patient information and no null patient ids allowed. "
                    "These samples are missing patient data: {}\n".format(
                        ", ".join(clinicaldf[sampleId][
                                  clinicaldf[patientId] == ""]))
                )
            # CHECK: All patients should have associated sample data
            if not all(clinicaldf[sampleId] != ""):
                # ## MAKE WARNING FOR NOW###
                warning.write(
                    "Sample Clinical File: All patients must have associated "
                    "sample information. These patients are missing sample "
                    "data: {}\n".format(
                        ", ".join(clinicaldf[patientId][
                                  clinicaldf[sampleId] == ""]))
                )

        # CHECK: AGE_AT_SEQ_REPORT
        age = "AGE_AT_SEQ_REPORT"
        haveColumn = process_functions.checkColExist(clinicaldf, age)
        if haveColumn:
            # Deal with HIPAA converted rows from DFCI
            # First for loop can't int(text) because there
            # are instances that have <3435
            age_seq_report_df = clinicaldf[
                ~clinicaldf[age].isin(['Unknown', '>32485', '<6570'])
            ]

            # age_seq_report_df[age] = \
            #     remove_greaterthan_lessthan_str(age_seq_report_df[age])

            if not all([process_functions.checkInt(i)
                        for i in age_seq_report_df[age]]):
                total_error.write(
                    "Sample Clinical File: Please double check your "
                    "AGE_AT_SEQ_REPORT. It must be an integer, 'Unknown', "
                    "'>32485', '<6570'.\n"
                )
            else:
                age_seq_report_df[age] = age_seq_report_df[age].astype(int)
                median_age = age_seq_report_df[age].median()
                if median_age < 100:
                    total_error.write(
                        "Sample Clinical File: Please double check your "
                        "AGE_AT_SEQ_REPORT. You may be reporting this value "
                        "in YEARS, please report in DAYS.\n"
                    )
        else:
            total_error.write(
                "Sample Clinical File: Must have AGE_AT_SEQ_REPORT column.\n"
            )

        # CHECK: ONCOTREE_CODE
        haveColumn = \
            process_functions.checkColExist(clinicaldf, "ONCOTREE_CODE")
        maleOncoCodes = ["TESTIS", "PROSTATE", "PENIS"]
        womenOncoCodes = ["CERVIX", "VULVA", "UTERUS", "OVARY"]
        if haveColumn:
            # Make oncotree codes uppercase (SpCC/SPCC)
            clinicaldf['ONCOTREE_CODE'] = \
                clinicaldf['ONCOTREE_CODE'].astype(str).str.upper()

            oncotree_codes = clinicaldf['ONCOTREE_CODE'][
                clinicaldf['ONCOTREE_CODE'] != "UNKNOWN"]

            if not all(oncotree_codes.isin(oncotree_mapping['ONCOTREE_CODE'])):
                unmapped_oncotrees = oncotree_codes[
                    ~oncotree_codes.isin(oncotree_mapping['ONCOTREE_CODE'])]
                total_error.write(
                    "Sample Clinical File: Please double check that all your "
                    "ONCOTREE CODES exist in the mapping. You have {} samples "
                    "that don't map. These are the codes that "
                    "don't map: {}\n".format(
                        len(unmapped_oncotrees),
                        ",".join(set(unmapped_oncotrees)))
                )
            # Should add the SEX mismatch into the dashboard file
            if process_functions.checkColExist(clinicaldf, "SEX") and \
               'oncotree_mapping_dict' in locals() and \
               havePatientColumn and \
               haveSampleColumn:

                wrongCodeSamples = []
                # This is to check if oncotree codes match the sex,
                # returns list of samples that have conflicting codes and sex
                for code, patient, sample in zip(
                        clinicaldf['ONCOTREE_CODE'],
                        clinicaldf['PATIENT_ID'],
                        clinicaldf['SAMPLE_ID']):

                    if oncotree_mapping_dict.get(code) is not None and \
                       sum(clinicaldf['PATIENT_ID'] == patient) > 0:

                        primaryCode = oncotree_mapping_dict[code][
                            'ONCOTREE_PRIMARY_NODE']

                        sex = clinicaldf['SEX'][
                            clinicaldf['PATIENT_ID'] == patient].values[0]
                        sex = float('nan') if sex == '' else float(sex)
                        if oncotree_mapping_dict[code][
                                'ONCOTREE_PRIMARY_NODE'] in maleOncoCodes and \
                           sex != 1.0:

                            wrongCodeSamples.append(sample)
                        if oncotree_mapping_dict[code][
                                'ONCOTREE_PRIMARY_NODE'] in womenOncoCodes and\
                           sex != 2.0:

                            wrongCodeSamples.append(sample)
                if len(wrongCodeSamples) > 0:
                    warning.write(
                        "Sample Clinical File: Some SAMPLE_IDs have "
                        "conflicting SEX and ONCOTREE_CODES: {}\n".format(
                            ",".join(wrongCodeSamples))
                    )
        else:
            total_error.write(
                "Sample Clinical File: Must have ONCOTREE_CODE column.\n"
            )

        warn, error = process_functions.check_col_and_values(
            clinicaldf, "SAMPLE_TYPE", sampletype_mapping['CODE'].tolist(),
            "Sample Clinical File", required=True)
        total_error.write(error)

        # CHECK: SEQ_ASSAY_ID
        haveColumn = process_functions.checkColExist(clinicaldf,
                                                     "SEQ_ASSAY_ID")
        if haveColumn:
            if not all([i != "" for i in clinicaldf['SEQ_ASSAY_ID']]):
                total_error.write(
                    "Sample Clinical File: Please double check your "
                    "SEQ_ASSAY_ID columns, there are empty rows.\n"
                )
            # must remove empty seq assay ids first
            # Checking if seq assay ids start with the center name
            empty_seq_idx = clinicaldf.SEQ_ASSAY_ID != ""
            seqassay_ids = clinicaldf.SEQ_ASSAY_ID[empty_seq_idx]
            uniq_seqassay_ids = seqassay_ids.unique()
            invalid_seqassay = []
            for seqassay in uniq_seqassay_ids:
                # SEQ Ids are all capitalized now, so no need to check
                # for differences in case
                if not seqassay.upper().startswith(self.center):
                    invalid_seqassay.append(seqassay)
            if invalid_seqassay:
                total_error.write(
                    "Sample Clinical File: Please make sure your "
                    "SEQ_ASSAY_IDs start with your center "
                    "abbreviation: {}.\n".format(
                        ", ".join(invalid_seqassay))
                )
        else:
            total_error.write(
                "Sample Clinical File: Must have SEQ_ASSAY_ID column.\n"
            )

        haveColumn = process_functions.checkColExist(clinicaldf, "SEQ_DATE")
        seq_date_error = (
            "Sample Clinical File: SEQ_DATE must be one of five values- "
            "For Jan-March: use Jan-YEAR. "
            "For Apr-June: use Apr-YEAR. "
            "For July-Sep: use Jul-YEAR. "
            "For Oct-Dec: use Oct-YEAR. (ie. Apr-2017) "
            "For values that don't have SEQ_DATES that "
            "you want released use 'release'.\n"
        )

        if haveColumn:
            clinicaldf['SEQ_DATE'] = [
                i.title() for i in clinicaldf['SEQ_DATE'].astype(str)
            ]

            seqdate = clinicaldf['SEQ_DATE'][
                clinicaldf['SEQ_DATE'] != 'Release']
            if sum(clinicaldf['SEQ_DATE'] == '') > 0:
                total_error.write(
                    "Sample Clinical File: Samples without SEQ_DATEs will "
                    "NOT be released.\n"
                )
            try:
                if not seqdate.empty:
                    seqdate.apply(
                        lambda date: datetime.datetime.strptime(date, '%b-%Y'))
                    if not seqdate.str.startswith(
                            ("Jan", "Apr", "Jul", "Oct")).all():
                        total_error.write(seq_date_error)
            except ValueError:
                total_error.write(seq_date_error)
        else:
            total_error.write(
                "Sample Clinical File: Must have SEQ_DATE column.\n"
            )

        # CHECK: BIRTH_YEAR
        error = _check_year(clinicaldf=clinicaldf,
                            year_col="BIRTH_YEAR",
                            filename="Patient Clinical File",
                            allowed_string_values=['Unknown', '>89', '<18'])
        total_error.write(error)

        # CHECK: YEAR DEATH
        error = _check_year(clinicaldf=clinicaldf,
                            year_col="YEAR_DEATH",
                            filename="Patient Clinical File",
                            allowed_string_values=['Unknown', 'Not Collected',
                                                   'Not Applicable',
                                                   "Not Released",
                                                   '>89', '<18'])
        total_error.write(error)

        # CHECK: YEAR CONTACT
        error = _check_year(clinicaldf=clinicaldf,
                            year_col="YEAR_CONTACT",
                            filename="Patient Clinical File",
                            allowed_string_values=['Unknown', 'Not Collected',
                                                   "Not Released",
                                                   '>89', '<18'])
        total_error.write(error)

        # CHECK: INT CONTACT
        haveColumn = process_functions.checkColExist(clinicaldf, "INT_CONTACT")
        if haveColumn:
            if not all([process_functions.checkInt(i)
                        for i in clinicaldf.INT_CONTACT if i not in
                        ['>32485', '<6570', 'Unknown', 'Not Collected',
                         "Not Released"]]):

                total_error.write(
                    "Patient Clinical File: Please double check your "
                    "INT_CONTACT column, it must be an integer, '>32485', "
                    "'<6570', 'Unknown', 'Not Released' or 'Not Collected'.\n"
                )
        else:
            total_error.write(
                "Patient Clinical File: Must have INT_CONTACT column.\n"
            )

        # INT DOD
        haveColumn = process_functions.checkColExist(clinicaldf, "INT_DOD")
        if haveColumn:
            if not all([process_functions.checkInt(i)
                        for i in clinicaldf.INT_DOD if i not in
                        ['>32485', '<6570', 'Unknown',
                         'Not Collected', 'Not Applicable', "Not Released"]]):

                total_error.write(
                    "Patient Clinical File: Please double check your INT_DOD "
                    "column, it must be an integer, '>32485', '<6570', "
                    "'Unknown', 'Not Collected', 'Not Released' or "
                    "'Not Applicable'.\n"
                )
        else:
            total_error.write(
                "Patient Clinical File: Must have INT_DOD column.\n"
            )

        haveColumn = process_functions.checkColExist(clinicaldf, "DEAD")
        if haveColumn:
            # Need to have check_bool function
            if not all([
                    str(i).upper() in ['TRUE', 'FALSE']
                    for i in clinicaldf.DEAD if i not in
                    ['Unknown', 'Not Collected', "Not Released"]]):
                total_error.write(
                    "Patient Clinical File: Please double check your "
                    "DEAD column, it must be True, False, 'Unknown', "
                    "'Not Released' or 'Not Collected'.\n"
                )
        else:
            total_error.write(
                "Patient Clinical File: Must have DEAD column.\n"
            )

        # CHECK: PRIMARY_RACE
        warn, error = process_functions.check_col_and_values(
            clinicaldf, "PRIMARY_RACE", race_mapping['CODE'].tolist(),
            "Patient Clinical File"
        )
        warning.write(warn)
        total_error.write(error)

        # CHECK: SECONDARY_RACE
        warn, error = process_functions.check_col_and_values(
            clinicaldf, "SECONDARY_RACE", race_mapping['CODE'].tolist(),
            "Patient Clinical File"
        )
        warning.write(warn)
        total_error.write(error)

        # CHECK: TERTIARY_RACE
        warn, error = process_functions.check_col_and_values(
            clinicaldf, "TERTIARY_RACE", race_mapping['CODE'].tolist(),
            "Patient Clinical File"
        )
        warning.write(warn)
        total_error.write(error)

        # CHECK: SEX
        warn, error = process_functions.check_col_and_values(
            clinicaldf, "SEX", sex_mapping['CODE'].tolist(),
            "Patient Clinical File", required=True
        )
        warning.write(warn)
        total_error.write(error)

        # CHECK: ETHNICITY
        warn, error = process_functions.check_col_and_values(
            clinicaldf, "ETHNICITY", ethnicity_mapping['CODE'].tolist(),
            "Patient Clinical File"
        )
        warning.write(warn)
        total_error.write(error)

        return total_error.getvalue(), warning.getvalue()

    def _get_dataframe(self, filePathList):
        clinicaldf = pd.read_csv(filePathList[0], sep="\t", comment="#")
        if len(filePathList) > 1:
            other_clinicaldf = pd.read_csv(filePathList[1], sep="\t",
                                           comment="#")
            try:
                clinicaldf = clinicaldf.merge(other_clinicaldf, on="PATIENT_ID")
            except Exception:
                raise ValueError((
                    "If submitting separate patient and sample files, "
                    "they both must have the PATIENT_ID column"))
            # Must figure out which is sample and which is patient
            if "sample" in filePathList[0]:
                sample = clinicaldf
                patient = other_clinicaldf
            else:
                sample = other_clinicaldf
                patient = clinicaldf

            if not all(sample['PATIENT_ID'].isin(patient['PATIENT_ID'])):
                raise ValueError((
                    "Patient Clinical File: All samples must have associated "
                    "patient information"))

        return clinicaldf



"""Test write invalid reasons module"""
from unittest import mock
from unittest.mock import create_autospec, Mock, patch

import genie
from genie import write_invalid_reasons
import pandas as pd
import synapseclient


SYN = create_autospec(synapseclient.Synapse)
CENTER_ERRORSDF = pd.DataFrame(
    {'id': ["syn1234", "syn2345"],
     'errors': ['my errors|n', "errors here|f"],
     'center': ['SAGE', 'TEST']}
)
ENT1 = synapseclient.File(id="syn123", name="test", parentId="syn3333")


class QueryResponse:
    def asDataFrame():
        return CENTER_ERRORSDF


def test__combine_center_file_errors():
    """Test combining each center's file errors"""
    expected_error = (
        f"\t{ENT1.name} ({ENT1.id}):\n\nmy errors\nn\n\n"
        f"\t{ENT1.name} ({ENT1.id}):\n\nerrors here\nf\n\n"
    )
    calls = [mock.call("syn1234", downloadFile=False),
             mock.call("syn2345", downloadFile=False)]
    with patch.object(SYN, "get", return_value=ENT1) as patch_synget:
        center_errors = write_invalid_reasons._combine_center_file_errors(
            SYN, CENTER_ERRORSDF
        )
        assert center_errors == expected_error
        patch_synget.assert_has_calls(calls)


def test_get_center_invalid_errors():
    """Test getting all center invalid errors"""
    with patch.object(SYN, "tableQuery",
                      return_value=QueryResponse) as patch_query,\
         patch.object(write_invalid_reasons, "_combine_center_file_errors",
                      return_value="errors") as patch_combine:
        center_invalid = write_invalid_reasons.get_center_invalid_errors(
           SYN, "syn3333"
        )
        assert center_invalid == {'SAGE': 'errors', 'TEST': 'errors'}
        patch_query.assert_called_once_with("SELECT * FROM syn3333")
        assert patch_combine.call_count == 2



#!/usr/bin/env python3
# noqa pylint: disable=line-too-long
"""genie cli"""
import argparse
import logging

import synapseclient

from . import create_case_lists, validate, write_invalid_reasons
from .__version__ import __version__


logger = logging.getLogger(__name__)


def synapse_login(username=None, password=None):
    """
    This function logs into synapse for you if credentials are saved.
    If not saved, then user is prompted username and password.

    :returns:     Synapseclient object
    """
    try:
        syn = synapseclient.login(silent=True)
    except Exception:
        if username is None and password is None:
            raise ValueError("Please specify --syn_user, --syn_pass to specify your Synapse "
                             "login. Please view https://docs.synapse.org/articles/client_configuration.html"
                             "to learn about logging into Synapse via the Python client.")
        syn = synapseclient.login(email=username,
                                  password=password,
                                  silent=True)
    return syn


def perform_create_case_list(syn, args):
    """CLI to create case lists given a clinical and assay information file"""
    create_case_lists.main(args.clinical_file_name,
                           args.assay_info_file_name,
                           args.output_dir,
                           args.study_id)


def perform_get_file_errors(syn, args):
    """CLI to get invalid reasons"""
    project = syn.get(args.project_id)
    db_mapping = syn.tableQuery(f"select * from {project.dbMapping[0]}")
    db_mappingdf = db_mapping.asDataFrame()
    error_tracker_synid = db_mappingdf['Id'][
        db_mappingdf['Database'] == "errorTracker"
    ][0]
    center_errors = write_invalid_reasons.get_center_invalid_errors(
        syn, error_tracker_synid
    )
    print(center_errors[args.center])


def build_parser():
    parser = argparse.ArgumentParser(description='GENIE processing')

    parser.add_argument("--syn_user", type=str, help='Synapse username')

    parser.add_argument("--syn_pass", type=str, help='Synapse password')

    parser.add_argument('-v', '--version', action='version',
                        version='genie {}'.format(__version__))

    subparsers = parser.add_subparsers(title='commands',
                                       description='The following commands are available:',
                                       help='For additional help: "genie <COMMAND> -h"')

    parser_validate = subparsers.add_parser('validate', help='Validates GENIE file formats')

    parser_validate.add_argument("filepath", type=str, nargs="+",
                                 help='File(s) that you are validating.'
                                      'If you validation your clinical files and you have both sample and '
                                      'patient files, you must provide both')

    parser_validate.add_argument("center", type=str, help='Contributing Centers')

    parser_validate.add_argument("--format_registry_packages", type=str, nargs="+",
                                 default=["genie_registry"],
                                 help="Python package name(s) to get valid file formats from (default: %(default)s).")

    parser_validate.add_argument("--oncotree_link", type=str, help="Link to oncotree code")

    validate_group = parser_validate.add_mutually_exclusive_group()

    validate_group.add_argument("--filetype", type=str,
                                help='By default, the validator uses the filename to match '
                                     'the file format.  If your filename is incorrectly named, '
                                     'it will be invalid.  If you know the file format you are '
                                     'validating, you can ignore the filename validation and skip '
                                     'to file content validation. '
                                     'Note, the filetypes with SP at '
                                     'the end are for special sponsored projects.')

    validate_group.add_argument("--parentid", type=str, default=None,
                                help='Synapse id of center input folder. '
                                     'If specified, your valid files will be uploaded '
                                     'to this directory.')

    # TODO: remove this default when private genie project is ready
    parser_validate.add_argument("--project_id", type=str,
                                 default="syn3380222",
                                 help='Synapse Project ID where data is stored. (default: %(default)s).')

    parser_validate.add_argument("--nosymbol-check", action='store_true',
                                 help='Do not check hugo symbols of fusion and cna file')

    parser_validate.set_defaults(func=validate._perform_validate)

    parser_create_case = subparsers.add_parser(
        'create-case-lists', help='Creates cBioPortal case list files'
    )
    parser_create_case.add_argument("clinical_file_name",
                                    type=str,
                                    help="Clinical file path")
    parser_create_case.add_argument("assay_info_file_name",
                                    type=str,
                                    help="gene matrix file path")
    parser_create_case.add_argument("output_dir",
                                    type=str,
                                    help="Output directory")
    parser_create_case.add_argument("study_id",
                                    type=str,
                                    help="Output directory")
    parser_create_case.set_defaults(func=perform_create_case_list)

    parser_get_invalid = subparsers.add_parser(
        'get-file-errors', help='Get the file invalid reasons for a specific center'
    )
    parser_get_invalid.add_argument("center", type=str, help='Contributing Centers')
    parser_get_invalid.add_argument(
        "--project_id", type=str,
        default="syn3380222",
        help='Synapse Project ID where data is stored. (default: %(default)s).'
    )
    parser_get_invalid.set_defaults(func=perform_get_file_errors)

    return parser


def main():
    """Invoke"""
    args = build_parser().parse_args()
    syn = synapse_login(args.syn_user, args.syn_pass)
    # func has to match the set_defaults
    args.func(syn, args)


if __name__ == "__main__":
    main()



"""Tests database to staging functions"""
import os
from unittest import mock
from unittest.mock import patch

import pandas as pd
import synapseclient

from genie import database_to_staging, process_functions

SYN = synapseclient.Synapse()
FILEVIEW_SYNID = "syn12345"
GENIE_VERSION = "vTEST"
CONSORTIUM_SYNID = "syn2222"


class Tablequerydf():
    """tableQuery.asDataFrame() class"""
    def __init__(self, df):
        self.df = df

    def asDataFrame(self):
        return self.df


def test_store_gene_panel_files():
    current_release_staging = False

    data_gene_panel = pd.DataFrame({'mutations': ['PANEL1']})
    gene_paneldf = pd.DataFrame({'id': ['syn3333']})

    with mock.patch.object(
            SYN, "tableQuery",
            return_value=Tablequerydf(gene_paneldf)) as patch_syn_table_query,\
         mock.patch.object(
             database_to_staging, "store_file",
             return_value=synapseclient.Entity()) as patch_storefile,\
         mock.patch.object(
             SYN, "get",
             return_value=synapseclient.Entity(
                 path="/foo/bar/PANEL1.txt",
                 versionNumber=2)) as patch_syn_get,\
         mock.patch.object(os, "rename") as patch_os_rename:

        database_to_staging.store_gene_panel_files(SYN,
                                                   FILEVIEW_SYNID,
                                                   GENIE_VERSION,
                                                   data_gene_panel,
                                                   CONSORTIUM_SYNID,
                                                   ["TEST"])

        patch_syn_table_query.assert_called_once_with(
            "select id from %s where cBioFileFormat = 'genePanel' "
            "and fileStage = 'staging' and "
            "name not in ('data_gene_panel_TEST.txt')" % FILEVIEW_SYNID)

        patch_storefile.assert_called_once_with(
            SYN,
            os.path.join(database_to_staging.GENIE_RELEASE_DIR,
                         "PANEL1_vTEST.txt"),
            parent=CONSORTIUM_SYNID,
            genieVersion=GENIE_VERSION,
            name="PANEL1.txt",
            cBioFileFormat="genePanel",
            used='syn3333.2')

        patch_syn_get.assert_called_once_with('syn3333')
        patch_os_rename.assert_called_once_with(
            "/foo/bar/PANEL1.txt",
            os.path.join(database_to_staging.GENIE_RELEASE_DIR,
                         "PANEL1_vTEST.txt"))


def test_store_assay_info_files():
    """Tests storing of assay information file"""
    assay_infodf = pd.DataFrame({'library_strategy': ['WXS'],
                                 'SEQ_ASSAY_ID': ['A']})
    clinicaldf = pd.DataFrame({'SEQ_ASSAY_ID': ['A']})
    database_to_staging.GENIE_RELEASE_DIR = "./"
    path = os.path.join(database_to_staging.GENIE_RELEASE_DIR,
                        "assay_information_vTEST.txt")
    with patch.object(SYN, "create_snapshot_version",
                      return_value=2) as patch_create_version,\
         patch.object(process_functions, "get_syntabledf",
                      return_value=assay_infodf) as patch_table_query,\
         patch.object(database_to_staging, "store_file",
                      return_value=synapseclient.Entity()) as patch_storefile:
        wes_ids = database_to_staging.store_assay_info_files(SYN,
                                                             GENIE_VERSION,
                                                             FILEVIEW_SYNID,
                                                             clinicaldf,
                                                             CONSORTIUM_SYNID)
        patch_create_version.assert_called_once_with(FILEVIEW_SYNID,
                                                     comment=GENIE_VERSION)
        patch_table_query.assert_called_once_with(
            SYN,
            f"select * from {FILEVIEW_SYNID} where SEQ_ASSAY_ID in ('A')"
        )
        patch_storefile.assert_called_once_with(SYN, path,
                                                parent=CONSORTIUM_SYNID,
                                                genieVersion=GENIE_VERSION,
                                                name="assay_information.txt",
                                                used=f"{FILEVIEW_SYNID}.2")
        assert wes_ids == ['A']
